@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2006-2014 PLT Design Inc.
+  Copyright (c) 2006-2018 PLT Design Inc.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -161,9 +161,9 @@ static void set_cache(void *p, Scheme_Object *last)
 
 Scheme_Object *scheme_native_stack_trace(void)
 {
-  void *p, *q, *cache_frame_p;
+  void *p, *q, *cache_frame_p, *set_cache_sp = NULL;
   uintptr_t stack_end, real_stack_end, stack_start, halfway;
-  Scheme_Object *name, *last = NULL, *first = NULL, *tail;
+  Scheme_Object *name, *last = NULL, *first = NULL, *tail, *set_cache_last = NULL;
   int cache_had_name = 0;
 #ifdef MZ_USE_DWARF_LIBUNWIND
   unw_context_t cx;
@@ -198,9 +198,10 @@ Scheme_Object *scheme_native_stack_trace(void)
   }
 
 #ifdef MZ_USE_DWARF_LIBUNWIND
-  unw_getcontext(&cx);
+  if (unw_getcontext(&cx) != 0)
+    return NULL;
   unw_init_local(&c, &cx);
-  unw_set_safe_pointer_range(&c, stack_start, stack_end);
+  unw_set_safe_pointer_range(&c, stack_start, real_stack_end);
   use_unw = 1;
   p = NULL;
 #else
@@ -289,14 +290,15 @@ Scheme_Object *scheme_native_stack_trace(void)
 	  first = name;
 	last = name;
 	if (shift_cache_to_next) {
-	  stack_cache_stack[stack_cache_stack_pos].cache = last;
+          set_cache_last = last;
 	  shift_cache_to_next = 0;
 	}
       }
 
       if (cache_sp) {
 	if (STK_COMP((uintptr_t)halfway, (uintptr_t)cache_sp)) {
-	  set_cache(cache_sp, last);
+          set_cache_sp = cache_sp;
+          set_cache_last = last;
 	  if (!name)
 	    shift_cache_to_next = 1;
 	  halfway = stack_end;
@@ -313,7 +315,10 @@ Scheme_Object *scheme_native_stack_trace(void)
     }
 
     if (shift_cache_to_next)
-      stack_cache_stack[stack_cache_stack_pos].cache = tail;
+      set_cache_last = tail;
+
+    if (set_cache_sp)
+      set_cache(set_cache_sp, set_cache_last);
 
     if (last)
       SCHEME_CDR(last) = tail;
@@ -438,7 +443,7 @@ Scheme_Object *scheme_native_stack_trace(void)
 	first = name;
       last = name;
       if (shift_cache_to_next) {
-        stack_cache_stack[stack_cache_stack_pos].cache = last;
+        set_cache_last = last;
         shift_cache_to_next = 0;
       }
       added_list_elem = 1;
@@ -458,7 +463,8 @@ Scheme_Object *scheme_native_stack_trace(void)
        it will use the return address from the stack. */
     if (STK_COMP((uintptr_t)halfway, (uintptr_t)cache_frame_p)
 	&& cache_had_name) {
-      set_cache(cache_frame_p, last);
+      set_cache_sp = cache_frame_p;
+      set_cache_last = last;
       if (!added_list_elem)
 	shift_cache_to_next = 1;
       halfway = stack_end;
@@ -509,7 +515,10 @@ Scheme_Object *scheme_native_stack_trace(void)
   }
 
   if (shift_cache_to_next)
-    stack_cache_stack[stack_cache_stack_pos].cache = tail;
+    set_cache_last = tail;
+
+  if (set_cache_sp)
+    set_cache(set_cache_sp, set_cache_last);
 
 #ifdef MZ_USE_DWARF_LIBUNWIND
   unw_destroy_local(&c);
@@ -676,7 +685,7 @@ void scheme_jit_now(Scheme_Object *f)
 {
   if (SAME_TYPE(SCHEME_TYPE(f), scheme_native_closure_type)) {
     Scheme_Native_Closure *nc;
-    Scheme_Native_Closure_Data *ncd;
+    Scheme_Native_Lambda *ncd;
 
     nc = (Scheme_Native_Closure*)f;
     ncd = nc->code;
@@ -686,39 +695,36 @@ void scheme_jit_now(Scheme_Object *f)
 }
 
 
-typedef void *(*Module_Run_Proc)(Scheme_Env *menv, Scheme_Env *env, Scheme_Object **name);
-typedef void *(*Module_Exprun_Proc)(Scheme_Env *menv, int set_ns, Scheme_Object **name);
-typedef void *(*Module_Start_Proc)(struct Start_Module_Args *a, Scheme_Object **name);
+typedef Scheme_Object *(*Linklet_Run_Proc)(Scheme_Linklet *linklet, Scheme_Instance *inst, Scheme_Object **name);
+typedef void (*Thread_Start_Child_Proc)(Scheme_Thread *child, Scheme_Object *child_thunk);
 
-void *scheme_module_run_start(Scheme_Env *menv, Scheme_Env *env, Scheme_Object *name)
+Scheme_Object *scheme_linklet_run_start(Scheme_Linklet *linklet, Scheme_Instance *inst, Scheme_Object *name)
 {
-  Module_Run_Proc proc = (Module_Run_Proc)sjc.module_run_start_code;
-  if (proc)
-    return proc(menv, env, &name);
+  Linklet_Run_Proc proc = (Linklet_Run_Proc)sjc.linklet_run_start_code;
+  if (proc && !CHECK_RUNSTACK_REGISTER_UPDATE)
+    return proc(linklet, inst, &name);
   else
-    return scheme_module_run_finish(menv, env);
+    return scheme_linklet_run_finish(linklet, inst, 1);
 }
 
-void *scheme_module_exprun_start(Scheme_Env *menv, int set_ns, Scheme_Object *name)
+void scheme_thread_start_child(Scheme_Thread *child, Scheme_Object *child_thunk)
+  XFORM_SKIP_PROC
 {
-  Module_Exprun_Proc proc = (Module_Exprun_Proc)sjc.module_exprun_start_code;
-  if (proc)
-    return proc(menv, set_ns, &name);
+  Thread_Start_Child_Proc proc = (Thread_Start_Child_Proc)sjc.thread_start_child_code;
+  if (proc && !CHECK_RUNSTACK_REGISTER_UPDATE)
+    proc(child, child_thunk);
   else
-    return scheme_module_exprun_finish(menv, set_ns);
+    scheme_do_thread_start_child(child, child_thunk);
 }
 
-void *scheme_module_start_start(struct Start_Module_Args *a, Scheme_Object *name)
-{
-  Module_Start_Proc proc = (Module_Start_Proc)sjc.module_start_start_code;
-  if (proc)
-    return proc(a, &name);
-  else
-    return scheme_module_start_finish(a);
-}
 
 #else
 
 void* scheme_jit_find_code_end(void *p) { return NULL; }
 
+void scheme_thread_start_child(Scheme_Thread *child, Scheme_Object *child_thunk)
+{
+  scheme_do_thread_start_child(child, child_thunk);
+}
+
 #endif
